from bs4 import BeautifulSoup
from ebooklib import ITEM_DOCUMENT, epub


class EPubReader:
    def __init__(self, filename):
        self.book = epub.read_epub(filename, {"ignore_ncx": True})

    def __iter__(self):
        self.it = self.book.get_items_of_type(ITEM_DOCUMENT)
        return self

    def __next__(self):
        try:
            item = next(self.it)
        except StopIteration:
            raise StopIteration

        soup = BeautifulSoup(item.get_body_content(), "html.parser")
        return soup.get_text()


if __name__ == "__main__":
    for item in EPubReader("workdir/InfiniteJest.epub"):
        print(len(item))



from xraygpt.db.chroma import ChromaDatabase


def printDatabase(filename):
    db = ChromaDatabase(None, filename + ".chroma")
    data = db.dump()
    for i in data:
        print(i["name"])
        print(i["description"])
        print("=" * 80)

    print("Total items:", len(data))


if __name__ == "__main__":
    printDatabase("workdir/InfiniteJest.epub")



import shelve

from loguru import logger

from xraygpt.db.chroma import ChromaDatabase
from xraygpt.llm import get_ebd, get_llm
from xraygpt.ner.agent import recognize_entities
from xraygpt.reader import EPubReader


def epubSummaryFlow(filename):
    people = set()
    for item in EPubReader(filename):
        logger.debug(item)
        people |= recognize_entities([item])
        logger.info(f"# people found so far: {len(people)}")


def epubPeopleFlow(filename):
    state = shelve.open(filename + ".shelve")
    llm = get_llm()
    ebd = get_ebd()
    db = ChromaDatabase(ebd, filename + ".chroma")

    for ix, item in enumerate(EPubReader(filename)):
        if ix <= state.get("last_processed", -1):
            logger.info(f"Skipping {ix}")
            continue
        # logger.debug(item)
        recognize_entities(item, llm, db)

        state["last_processed"] = ix

    for i in db.dump():
        print(i["name"])
        print(i["description"])
        print("=" * 80)


if __name__ == "__main__":
    print(epubPeopleFlow("workdir/InfiniteJest.epub"))



import argparse
import asyncio
import logging
import sys

from loguru import logger

from xraygpt.__about__ import __version__
from xraygpt.flow import epubPeopleFlow
from xraygpt.output import printDatabase

flowname_map = {
    "people": epubPeopleFlow,
    "peak": printDatabase,
}


def config_log_level(v: int) -> None:
    logger.remove()
    log_format = "<level>{message}</level>"
    if v >= 4:
        logger.add(sys.stderr, level="TRACE", format=log_format)
    elif v >= 3:
        logger.add(sys.stderr, level="DEBUG", format=log_format)
    elif v >= 2:
        logger.add(sys.stderr, level="INFO", format=log_format)
    elif v >= 1:
        logger.add(sys.stderr, level="SUCCESS", format=log_format)
    else:
        logger.add(sys.stderr, level="WARNING", format=log_format)

    if v < 4:
        # avoid "WARNING! deployment_id is not default parameter."
        langchain_logger = logging.getLogger("langchain.chat_models.openai")
        langchain_logger.disabled = True


def main() -> None:
    parser = argparse.ArgumentParser(description="xxy-" + __version__)
    parser.add_argument("-v", action="count", default=0, help="verbose level.")
    parser.add_argument("filename", help="input file name.")
    parser.add_argument(
        "-f", help="flow name", choices=flowname_map.keys(), default="people"
    )

    # create the parser for the "foo" command
    args = parser.parse_args()

    config_log_level(args.v)

    flowname_map[args.f](args.filename)


if __name__ == "__main__":
    main()



from typing import List, Optional, Tuple

from langchain.output_parsers import ResponseSchema, StructuredOutputParser
from langchain.prompts import (
    ChatPromptTemplate,
    HumanMessagePromptTemplate,
    SystemMessagePromptTemplate,
)
from langchain_core.messages import AIMessage, SystemMessage
from langchain_openai import ChatOpenAI
from loguru import logger

from xraygpt.db.base import Item


def _gross_recognize_entities(text: str, llm: ChatOpenAI) -> List[str]:
    # Define the prompt with a structured JSON schema
    response_schemas = [
        ResponseSchema(
            name="item", description="An array of entity names", type="[string]"
        ),
    ]

    output_parser = StructuredOutputParser.from_response_schemas(response_schemas)

    chat_template = ChatPromptTemplate(
        messages=[
            SystemMessagePromptTemplate.from_template(
                "You are an agent to help me recognize named entities in text, give me the entities based on following rules\n1. Entities are only people's names\n2. Outputs should following the format: {format_instructions}"
            ),
            HumanMessagePromptTemplate.from_template("{text}"),
        ],
        input_variables=["text"],
        partial_variables={
            "format_instructions": output_parser.get_format_instructions()
        },
    )

    chain = chat_template | llm | output_parser

    resp = chain.invoke({"text": text})
    logger.debug("{num_items} items recognized grossly", num_items=len(resp["item"]))
    return resp["item"]


def _refine_recognized_entity(
    text: str, name: str, items: List[Item], llm
) -> Tuple[List[str], Optional[Item]]:
    logger.debug(
        "Refining recognized entity: {name} with {num_items}",
        name=name,
        num_items=len(items),
    )
    response_schemas = [
        ResponseSchema(
            name="to_delete",
            description="An array of entity ids to delete",
            type="[int]",
        ),
        ResponseSchema(
            name="entity_name",
            description="Array of entity names of single entity",
            type="[string]",
        ),
        ResponseSchema(
            name="entity_description", description="entity description", type="string"
        ),
    ]

    output_parser = StructuredOutputParser.from_response_schemas(response_schemas)

    chat_template = ChatPromptTemplate(
        messages=[
            SystemMessagePromptTemplate.from_template(
                'You are an assistant tasked with refining a specific entity from text. Given the recognized entity "{name}" and related database information, your tasks are:\n1. Identify if the entity "{name}" is incorrect, irrelevant, or outdated, and mark it for deletion if necessary.\n2. Update the entity by deleting and adding a new entry if the text provides more important or detailed information.\n3. Ensure the updated entity name remains accurate.\n4. Each related entitys are marked with an ID in []. To delete provide the ID in "to_delete"\n5. Provide a simple and concise entity description with less than 100 words.\\nOnly process and output information for the entity "{name}".\nYour output must follow this format: {format_instructions}'
            ),
            HumanMessagePromptTemplate.from_template("Existing entity: {reference}"),
            HumanMessagePromptTemplate.from_template("{text}"),
        ],
        input_variables=["text", "name", "reference"],
        partial_variables={
            "format_instructions": output_parser.get_format_instructions()
        },
    )

    chain = chat_template | llm | output_parser

    references = [(",".join(i["name"]), i["description"]) for i in items]
    reference_description = "\n".join(
        [f"[{ix}]: {n}: {d}" for ix, (n, d) in enumerate(references)]
    )
    resp = chain.invoke(
        {"text": text, "name": name, "reference": reference_description}
    )
    logger.trace(
        "Reference description: {reference_description}",
        reference_description=reference_description,
    )
    logger.debug("{num_items} to delete", num_items=len(resp["to_delete"]))
    to_delete = [i for i in resp["to_delete"] if i < len(items)]
    if len(to_delete) != len(resp["to_delete"]):
        logger.warning(
            "Some items to delete are invalid: {to_delete}", to_delete=resp["to_delete"]
        )
    item_to_delete = [items[i]["id"] for i in to_delete]
    name_to_delete = [items[i]["name"] for i in to_delete]
    logger.debug("Items to delete: {name_to_delete}", name_to_delete=name_to_delete)

    item_to_add = None
    if resp["entity_name"]:
        item_to_add = Item(
            id="", name=resp["entity_name"], description=resp["entity_description"]
        )
    else:
        logger.warning("No item to add")

    return item_to_delete, item_to_add


def recognize_entities(text: str, llm: ChatOpenAI, db):
    items = _gross_recognize_entities(text, llm)
    for i in items:
        related = db.query(i)
        to_delete, to_add = _refine_recognized_entity(text, i, related, llm)
        for d in to_delete:
            db.delete(Item(id=d, name=[], description=""))

        if to_add:
            db.add(to_add)



import os
import re
from typing import Dict, Generator, List

from azure.ai.textanalytics import (
    DocumentError,
    RecognizeEntitiesResult,
    TextAnalyticsClient,
)
from azure.core.credentials import AzureKeyCredential
from loguru import logger

from xraygpt.splitter import sillySplit


def _recognize_entities(
    text_analytics_client: TextAnalyticsClient, reviews: List[str]
) -> Generator[RecognizeEntitiesResult | DocumentError, None, None]:
    for r in sillySplit(reviews, 5000):
        yield text_analytics_client.recognize_entities([r])[0]


def recognize_entities(reviews: List[str]) -> set[str]:
    endpoint = os.environ["AZURE_LANGUAGE_ENDPOINT"]
    key = os.environ["AZURE_LANGUAGE_KEY"]
    persons = set()

    text_analytics_client = TextAnalyticsClient(
        endpoint=endpoint, credential=AzureKeyCredential(key)
    )

    persons = set()

    for review in _recognize_entities(text_analytics_client, reviews):
        if review.is_error:
            logger.error("Error: {}", review["error"])
            continue

        for entity in review.entities:
            if entity.category == "Person" and entity.text not in persons:
                logger.debug("Person: {}", entity.text)
                persons.add(entity.text)

    return persons


if __name__ == "__main__":
    reviews = [
        """I work for Foo Company, and we hired Contoso for our annual founding ceremony. The food
        was amazing and we all can't say enough good words about the quality and the level of service.""",
        """We at the Foo Company re-hired Contoso after all of our past successes with the company.
        Though the food was still great, I feel there has been a quality drop since their last time
        catering for us. Is anyone else running into the same problem?""",
        """Bar Company is over the moon about the service we received from Contoso, the best sliders ever!!!!""",
    ]
    recognize_entities(reviews)



import itertools
from typing import Generator, List


def sillySplit(strs: List[str], max_len) -> Generator[str, None, None]:
    res = ""
    for i in itertools.chain.from_iterable([s.split(".") for s in strs]):
        if len(res) + len(i) + 1 > max_len:
            yield res
            res = ""
        res += i + "."



from typing import List, Optional
from uuid import uuid4

import chromadb
from langchain_openai import OpenAIEmbeddings
from loguru import logger

from xraygpt.db.base import Database, Item

SPLITTER = "|"


class ChromaDatabase(Database):
    def __init__(self, llm: OpenAIEmbeddings, path: Optional[str] = None):
        self.llm = llm
        if path is not None:
            client = chromadb.PersistentClient(path=path)
        else:
            client = chromadb.Client()
        self.collection = client.get_or_create_collection("people")

    def add(self, item: Item):
        new_id = uuid4().hex
        keys = SPLITTER.join(item["name"])
        logger.info("Adding item {name} with id {id}", name=keys, id=new_id)
        embedding = self.llm.embed_query(item["description"])
        self.collection.add(
            documents=[item["description"]],
            embeddings=[embedding],
            metadatas=[{"keys": keys}],
            ids=[new_id],
        )

    def delete(self, item: Item):
        logger.info("Deleting item with id {id}", id=item["id"])
        self.collection.delete(ids=[item["id"]])

    def query(self, name: str, n=3) -> List[Item]:
        embedding = self.llm.embed_query(name)
        results = self.collection.query(embedding, n_results=n)
        return [
            Item(id=ix, name=meta["keys"].split(SPLITTER), description=doc)
            for ix, doc, meta in zip(
                results["ids"][0], results["documents"][0], results["metadatas"][0]
            )
        ]

    def dump(self) -> List[Item]:
        results = self.collection.get(include=["documents", "metadatas"])
        return [
            Item(id=ix, name=meta["keys"].split(SPLITTER), description=doc)
            for ix, doc, meta in zip(
                results["ids"], results["documents"], results["metadatas"]
            )
        ]



from abc import abstractmethod
from typing import List, TypedDict


class Item(TypedDict):
    id: str
    name: List[str]
    description: str


class Database:
    @abstractmethod
    def add(self, item: Item) -> None:
        pass

    @abstractmethod
    def delete(self, item: Item) -> None:
        pass

    @abstractmethod
    def query(self, name: str) -> List[Item]:
        pass

    @abstractmethod
    def dump(self) -> List[Item]:
        pass
